require 'spec_helper'

RSpec.describe(KarrierWave::Mount) do
  let(:parent) { Class.new(ActiveRecord::Base) { self.table_name = 'messages' } }
  let(:first_child) { Class.new(parent) }
  let(:second_child) { Class.new(parent) }

  describe '#kage_options' do
    it { expect(first_child.kage_options.object_id).not_to eq(second_child.kage_options.object_id) }

    context 'when parent class has initialized kage_options' do
      let(:options) { { foo: 'bar' } }

      before { parent.kage_options.merge!(options) }

      it { expect(parent.kage_options).to eq(options) }
      it { expect(first_child.kage_options).to eq(parent.kage_options) }
      it { expect(first_child.kage_options).to eq(second_child.kage_options) }
      it { expect(first_child.kage_options.object_id).not_to eq(second_child.kage_options.object_id) }
      it { expect(first_child.kage_options.object_id).not_to eq(parent.kage_options.object_id) }
    end
  end

  describe '#mount_kage' do
    before { parent.new }

    shared_examples 'mountable' do
      let(:parent_instance) { parent.new }

      describe '#column_name' do
        it { expect(parent_instance.send(column_name)).to be_kind_of(KarrierWave::Mount::Mounter) }
        it { expect(parent_instance.send(column_name)).not_to eq(parent.new.send(column_name)) }
      end

      describe '#column_name=' do
        let(:new_file) { 'newfile' }

        it { expect(parent_instance.send(column_name)).to receive(:cache).with(new_file); parent_instance.send(:"#{column_name}=", new_file) }
      end

      describe '#remote_column_name_url' do
        it { expect(parent_instance.send(column_name)).to receive(:remote_url); parent_instance.send(:"remote_#{column_name}_url") }
      end

      describe '#remote_column_name_url=' do
        let(:url) { 'https://www.google.com/images/srpr/logo11w.png' }

        it { expect(parent_instance.send(column_name)).to receive(:remote_url=).with(url); parent_instance.send(:"remote_#{column_name}_url=", url) }
      end

      describe '#column_name?' do
        it { expect(parent_instance.send(column_name)).to receive(:present?); parent_instance.send(:"#{column_name}?") }
      end

      describe '#store_column_name!' do
        it { expect(parent_instance.send(column_name)).to receive(:store!); parent_instance.send(:"store_#{column_name}!") }

        context 'when invoked with model callback' do
          it { expect(parent_instance.send(column_name)).to receive(:store!); parent_instance.save }
        end

        context 'when invoked with model callback with dirty attributes' do
          let(:file) { File.open('spec/support/1x1.jpg') }

          before { parent_instance.send(:"#{column_name}=", file) }

          it { expect(parent_instance.send(column_name)).to receive(:store!); parent_instance.save }
        end
      end

      describe '#write_column_name_meta' do
        it { expect(parent_instance.send(column_name)).to receive(:write_meta); parent_instance.send(:"write_#{column_name}_meta") }

        context 'when invoked with model callback' do
          it { expect(parent_instance.send(column_name)).to receive(:write_meta); parent_instance.save }
        end

        context 'when invoked with model callback with dirty attributes' do
          let(:file) { File.open('spec/support/1x1.jpg') }

          before { parent_instance.send(:"#{column_name}=", file) }

          it { expect(parent_instance.send(column_name)).to receive(:write_meta); parent_instance.save }
        end
      end

      describe '#remove_column_name' do
        it { expect(parent_instance.send(column_name)).to receive(:remove!); parent_instance.send(:"remove_#{column_name}") }
        it { parent_instance.send(:"remove_#{column_name}"); expect(parent_instance.attachment_changed?).to eq(true) }

        context 'when invoked with model callback' do
          before { parent_instance.save }

          it { expect(parent_instance.send(column_name)).to receive(:remove!); parent_instance.destroy }
        end
      end
    end

    context 'when mount_on option is not provided' do
      it { expect(parent).to receive(:serialize).with(:attachment); parent.mount_kage(:attachment) }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.included_modules } }
      it { expect { parent.mount_kage(:attachment) }.not_to change { parent.new.respond_to?(:attachment) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.new.attachment } }
      it { expect { parent.mount_kage(:attachment) }.not_to change { parent.new.respond_to?(:attachment=) } }
      it { expect { parent.mount_kage(:attachment) }.not_to change { parent.new.respond_to?(:attachment?) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.new.respond_to?(:store_attachment!) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.new.respond_to?(:write_attachment_meta) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.new.respond_to?(:remove_attachment) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.new.respond_to?(:_mounter, true) } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.kage_options } }
      it { expect { parent.mount_kage(:attachment) }.to change { parent.kage_options[:attachment] } }

      context 'after mounted' do
        let(:column_name) { :attachment }

        before { parent.mount_kage(:attachment) }

        it_behaves_like 'mountable'
      end
    end

    context 'when mount_on option is provided' do
      it { expect(parent).to receive(:serialize).with(:attachment); parent.mount_kage(:photo, mount_on: :attachment) }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.included_modules } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:photo) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:photo=) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:photo?) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:store_photo!) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:write_photo_meta) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:remove_photo) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.new.respond_to?(:_mounter, true) } }
      it { expect { parent.mount_kage(:photo, mount_on: :attachment) }.to change { parent.kage_options[:photo] } }

      context 'after mounted' do
        let(:column_name) { :photo }

        before { parent.mount_kage(:photo, mount_on: :attachment) }

        it_behaves_like 'mountable'
      end
    end
  end

  describe(KarrierWave::Mount::Mounter) do
    describe '#cache' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }

      before { parent.mount_kage(:attachment) }

      context 'when setting new file' do
        let(:file) { File.open('spec/support/1x1.jpg') }

        it { mounter.cache(file); expect(instance.attachment_changed?).to eq(true) }
        it { expect { mounter.cache(file) }.not_to change { mounter.remove? } }
      end

      context 'when setting nil' do
        let(:file) { nil }

        it { mounter.cache(file); expect(instance.attachment_changed?).to eq(true) }
        it { expect { mounter.cache(file) }.to change { mounter.remove? } }
      end
    end

    describe '#store!' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }

      before { parent.mount_kage(:attachment) }

      context 'when nothing is set as cache' do
        it { expect(mounter.kage_file.send(:client)).not_to receive(:store); mounter.store! }
        it { mounter.store!; expect(mounter.send(:will_update?)).to eq(false) }
      end

      context 'when non-nil file is set as cache' do
        let(:file) { File.open('spec/support/1x1.jpg') }
        let(:identifier) { 'identifier' }

        before { mounter.cache(file) }
        before { allow(mounter.kage_file.send(:client)).to receive(:store) { identifier } }

        it { expect(mounter.kage_file.send(:client)).to receive(:store) { identifier }; mounter.store! }
        it { mounter.store!; expect(mounter.send(:will_update?)).to eq(true) }
      end

      context 'when nil is set as cache' do
        before { mounter.cache(nil) }

        it { expect(mounter.kage_file.send(:client)).not_to receive(:store); mounter.store! }
        it { mounter.store!; expect(mounter.send(:will_update?)).to eq(true) }
        it { expect(mounter.kage_file).to receive(:delete); mounter.store! }
      end
    end

    describe '#write_meta' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }
      let(:file) { mounter.kage_file }

      before { parent.mount_kage(:attachment) }

      context 'when it is needed to update' do
        before { mounter.send(:will_update!) }

        it { expect(file).to receive(:to_json) { {} }; mounter.write_meta }
      end

      context 'when it is not needed to update' do
        it { expect(file).not_to receive(:to_json); mounter.write_meta }
      end
    end

    describe '#remote_url=' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }
      let(:url) { 'https://www.google.com/images/srpr/logo11w.png' }
      let(:path) { '/path/to/tempfile' }
      let(:base_uri) { URI.parse(url) }
      let(:mock_tempfile) { double(path: path, meta: {}, base_uri: base_uri) }

      before { allow_any_instance_of(URI::HTTPS).to receive(:open) { mock_tempfile } }
      before { parent.mount_kage(:attachment) }

      it { expect { mounter.remote_url = url }.to change { mounter.remote_url } }
      it { mounter.remote_url = url; expect(instance.attachment_changed?).to eq(true) }
    end

    describe '#default_url' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }

      before { parent.mount_kage(:attachment, options) }

      context 'when it is string' do
        let(:default_fallback) { 'default.png' }
        let(:options) { { default_url: default_fallback } }
        let(:action_controller_base_stub) { double(helpers: stubbed_helpers) }
        let(:stubbed_helpers) { double() }

        before { allow(stubbed_helpers).to receive(:asset_path) { |asset| "assets/#{asset}" } }
        before { stub_const('ActionController::Base', action_controller_base_stub) }

        it { expect(mounter.default_url).to eq("assets/#{default_fallback}") }
      end

      context 'when it is array' do
        let(:default_fallbacks) { ['default.png', 'default1.png'] }
        let(:options) { { default_url: default_fallbacks } }
        let(:action_controller_base_stub) { double(helpers: stubbed_helpers) }
        let(:stubbed_helpers) { double() }

        before { allow(stubbed_helpers).to receive(:asset_path) { |asset| "assets/#{asset}" } }
        before { stub_const('ActionController::Base', action_controller_base_stub) }

        context 'when id is odd' do
          before { allow(instance).to receive(:id) { 1 } }

          it { expect(mounter.default_url).to eq("assets/#{default_fallbacks[1]}") }
        end

        context 'when id is even' do
          before { allow(instance).to receive(:id) { 2 } }

          it { expect(mounter.default_url).to eq("assets/#{default_fallbacks[0]}") }
        end
      end

      context 'when it is hash' do
        let(:default_fallbacks) { { small: 'default-small.png', large: 'default-large.png', default: 'default.png' } }
        let(:options) { { default_url: default_fallbacks } }
        let(:action_controller_base_stub) { double(helpers: stubbed_helpers) }
        let(:stubbed_helpers) { double() }

        before { allow(stubbed_helpers).to receive(:asset_path) { |asset| "assets/#{asset}" } }
        before { stub_const('ActionController::Base', action_controller_base_stub) }

        context 'when version is given' do
          it { expect(mounter.default_url(:large)).to eq("assets/#{default_fallbacks[:large]}") }
        end

        context 'when version is not given' do
          it { expect(mounter.default_url).to eq("assets/#{default_fallbacks[:default]}") }
        end
      end
    end

    describe '#remote_object_id=' do
      let(:instance) { parent.new }
      let(:mounter) { instance.attachment }
      let(:object_id) { 'a/b/c' }
      let(:new_access_key) { 'd/e/f' }

      before { allow_any_instance_of(KarrierWave::Client).to receive(:copy) { { access_key: new_access_key }.to_json } }
      before { allow_any_instance_of(KarrierWave::Client).to receive(:meta) { { original: { width: 100, height: 100 }, }.to_json } }
      before { parent.mount_kage(:attachment) }

      it { expect { mounter.remote_object_id = object_id }.to change { mounter.remote_object_id } }
      it { mounter.remote_object_id = object_id; expect(mounter.store!).to eq(true) }
      it { mounter.remote_object_id = object_id; mounter.store!; expect { mounter.write_meta }.to change { instance.send(:read_attribute, :attachment) } }
      it { mounter.remote_object_id = object_id; mounter.store!; expect { mounter.write_meta }.to change { instance.send(:read_attribute, :attachment) } }
      it { mounter.remote_object_id = object_id; expect(instance.attachment_changed?).to eq(true) }
    end
  end
end
